import asyncio
import os
from datetime import datetime
from typing import get_args, Union, Callable, Literal, List, Dict

import aiohttp
import rich_click as click
from rich.status import Status

from ._main import (
    reddit,
    Post,
    Posts,
    Search,
    Subreddit,
    Subreddits,
    User,
    Users,
)
from .meta.about import Project
from .meta.license import License
from .meta.version import Version
from .utils.general import General
from .utils.package import Package
from .utils.terminal import console, Message, Style

__all__ = ["start"]

package = Package(
    name=Project.package, version=Version, requester=reddit.connection.send_request
)


def help_callback(ctx: click.Context, option: click.Option, value: bool):
    """
    Custom callback function for handling the '--help' option in Click commands.

    Additionally, if the application is running as a Snap package, the
    function will pause execution, prompting the user to press any key
    before continuing. This is useful for when the user clicks the Knew Karma icon in application menu.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param option: The Click option that triggered this callback.
    :type option: click.Option
    :param value: The value of the custom help option. If True, the help
            message is displayed and the command execution is halted.
    :type value: bool
    """

    if value and not ctx.resilient_parsing:
        click.echo(ctx.get_help())
        if package.is_snap_package():
            click.pause()
        ctx.exit()


@click.command(
    name="license", help="Use this command to get licen[cs]e related information."
)
@click.option("-c", "--conditions", is_flag=True, help="Get licen[cs]e warranty.")
@click.option("-w", "--warranty", is_flag=True, help="Get licen[cs]e conditions.")
@click.pass_context
def license(ctx: click.Context, conditions: bool, warranty: bool):
    """
    Callback function for the `license` command.
    """
    if conditions:
        console.print(License.conditions, justify="center")
    elif warranty:
        console.print(License.warranty, justify="center")
    else:
        click.echo(ctx.command.get_usage(ctx=ctx))


@click.command(name="updates", help="Use this command to check for or install updates.")
@click.option("-c", "--check", is_flag=True, help="Check for updates.")
@click.option(
    "-i",
    "--install",
    is_flag=True,
    help="Install updates, if any are available [Non-Snap Packages only]",
)
@click.pass_context
def updates(ctx: click.Context, check: bool, install: bool):
    """
    Check for, or install updates if any are available.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param check: Option to check for updates.
    :type check: bool
    :param install: Option to install updates.
    :type install: bool
    """
    method_map = {
        "check": lambda session, status=None, message=None: package.check_updates(
            session=session, status=status
        ),
        "install": lambda session, status=None, message=None: package.check_updates(
            session=session, status=status, message=message, install_if_available=True
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=None,
            check=check,
            install=install,
        )
    )


@click.group(
    help=f"""
{Project.summary}
    
    
{Project.description}""",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "--export",
    type=str,
    help="A comma-separated list (no whitespaces) of file types to export the output to <supported: csv,html,json,xml>",
)
@click.option(
    "--limit",
    default=100,
    show_default=True,
    type=int,
    help=f"<bulk/semi-bulk> Maximum data output limit",
)
@click.option(
    "--sort",
    default="all",
    show_default=True,
    type=click.Choice(get_args(reddit.SORT)),
    help=f"<bulk/semi-bulk> Sort criterion",
)
@click.option(
    "--timeframe",
    default="all",
    show_default=True,
    type=click.Choice(get_args(reddit.TIMEFRAME)),
    help=f"<bulk/semi-bulk> Timeframe to get data from",
)
@click.option(
    "--time-format",
    default="locale",
    show_default=True,
    type=click.Choice(["concise", "locale"]),
    help=f"Determines the format of the output time",
)
@click.option(
    "-h",
    "--help",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=help_callback,
    help="Show this message and exit.",
)
@click.version_option(
    Version.full_version,
    "-v",
    "--version",
)
@click.pass_context
def cli(
    ctx: click.Context,
    timeframe: reddit.TIMEFRAME,
    sort: reddit.SORT,
    limit: int,
    time_format: str,
    export: List[General.EXPORT_FORMATS],
):
    """
    Main CLI group for Knew Karma.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param timeframe: Option to set the timeframe for the data.
    :type timeframe: Literal[str]
    :param sort: Option to set the sort criterion for the data.
    :type sort: Literal[str]
    :param limit: Option to set output data limit.
    :type limit: int
    :param time_format: Option to set the output's time format.
    :type time_format: Literal[str]
    :param export: Option to set data export file types.
    :type export: Literal[str]
    """

    ctx.ensure_object(Dict)
    ctx.obj["timeframe"] = timeframe
    ctx.obj["sort"] = sort
    ctx.obj["limit"] = limit
    ctx.obj["time_format"] = time_format
    ctx.obj["export"] = export


cli.add_command(cmd=license, name="license")
cli.add_command(cmd=updates, name="updates")


@cli.command(
    help="Use this command to get an individual post's data including its comments, "
    "provided the post's <id> and source <subreddit> are specified.",
)
@click.argument("id")
@click.argument("subreddit")
@click.option("--data", is_flag=True, help="Get post data")
@click.option("--comments", is_flag=True, help="Get post comments")
@click.pass_context
def post(ctx: click.Context, id: str, subreddit: str, data: bool, comments: bool):
    """
    Retrieve an individual post's data or comments.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param id: The ID of the post.
    :type id: str
    :param subreddit: The source subreddit of the post.
    :type subreddit: str
    :param data: Flag to get post data.
    :type data: bool
    :param comments: Flag to get post comments.
    :type comments: bool
    """

    sort: SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    post_instance = Post(id=id, subreddit=subreddit)
    method_map: Dict = {
        "comments": lambda session, status=None, message=None: post_instance.comments(
            limit=limit, sort=sort, status=status, message=message, session=session
        ),
        "data": lambda session, status=None, message=None: post_instance.data(
            session=session, status=status
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            data=data,
            comments=comments,
        )
    )


@cli.command(
    help="Use this command get best, controversial, front-page, new, popular, and/or rising posts.",
)
@click.option("--best", is_flag=True, help="Get posts from the best listing")
@click.option(
    "--controversial",
    is_flag=True,
    help="Get posts from the controversial listing",
)
@click.option(
    "--front-page",
    is_flag=True,
    help="Get posts from the reddit front-page",
)
@click.option("--new", is_flag=True, help="Get new posts")
@click.option(
    "--popular",
    is_flag=True,
    help="Get posts from the popular listing",
)
@click.option("--rising", is_flag=True, help="Get posts from the rising listing")
@click.pass_context
def posts(
    ctx: click.Context,
    best: bool,
    controversial: bool,
    front_page: bool,
    new: bool,
    popular: bool,
    rising: bool,
):
    """
    Retrieve various types of posts such as best, controversial, popular, new, and front-page.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param best: Flag to get posts from the best listing.
    :type best: bool
    :param controversial: Flag to get posts from the controversial listing.
    :type controversial: bool
    :param front_page: Flag to get posts from the reddit front-page.
    :type front_page: bool
    :param new: Flag to get new posts.
    :type new: bool
    :param popular: Flag to get posts from the popular listing.
    :type popular: bool
    :param rising: Flag to get posts from the rising listing.
    :type rising: bool
    """

    timeframe: reddit.TIMEFRAME = ctx.obj["timeframe"]
    sort: reddit.SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    posts_instance = Posts()
    method_map: Dict = {
        "best": lambda session, status=None, message=None: posts_instance.best(
            timeframe=timeframe,
            limit=limit,
            status=status,
            message=message,
            session=session,
        ),
        "controversial": lambda session, status=None, message=None: posts_instance.controversial(
            timeframe=timeframe,
            limit=limit,
            status=status,
            message=message,
            session=session,
        ),
        "front_page": lambda session, status=None, message=None: posts_instance.front_page(
            limit=limit, sort=sort, status=status, message=message, session=session
        ),
        "new": lambda session, status=None, message=None: posts_instance.new(
            limit=limit, sort=sort, status=status, message=message, session=session
        ),
        "popular": lambda session, status=None, message=None: posts_instance.popular(
            timeframe=timeframe,
            limit=limit,
            status=status,
            message=message,
            session=session,
        ),
        "rising": lambda session, status=None, message=None: posts_instance.rising(
            limit=limit, status=status, message=message, session=session
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            best=best,
            controversial=controversial,
            front_page=front_page,
            new=new,
            popular=popular,
            rising=rising,
        )
    )


@cli.command(
    help="Use this command for search/discovery of users, subreddits, and posts.",
)
@click.argument("query")
@click.option("--posts", is_flag=True, help="Search posts")
@click.option("--subreddits", is_flag=True, help="Search subreddits")
@click.option("--users", is_flag=True, help="Search users")
@click.pass_context
def search(ctx: click.Context, query: str, posts: bool, subreddits: bool, users: bool):
    """
    Search for posts, subreddits, or users based on a query.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param query: The search query.
    :type query: str
    :param posts: Flag to search posts.
    :type posts: bool
    :param subreddits: Flag to search subreddits.
    :type subreddits: bool
    :param users: Flag to search users.
    :type users: bool
    """

    sort: reddit.SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    search_instance = Search(
        query=query,
    )
    method_map: Dict = {
        "posts": lambda session, status=None, message=None: search_instance.posts(
            sort=sort, limit=limit, status=status, message=message, session=session
        ),
        "subreddits": lambda session, status=None, message=None: search_instance.subreddits(
            sort=sort, limit=limit, session=session
        ),
        "users": lambda session, status=None, message=None: search_instance.users(
            sort=sort, limit=limit, status=status, message=message, session=session
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            posts=posts,
            subreddits=subreddits,
            users=users,
        )
    )


@cli.command(
    help="Use this command to get a subreddit's data, such as comments, posts, wiki-pages, wiki-page data, and more...",
)
@click.argument("subreddit_name")
@click.option(
    "--comments",
    is_flag=True,
    help="Get a subreddit's comments (beta)",
)
@click.option(
    "--comments-per-post",
    type=int,
    help="To be used when getting comments with `--comments`",
)
@click.option("--profile", is_flag=True, help="Get a subreddit's profile")
@click.option("--posts", is_flag=True, help="Get a subreddit's posts")
@click.option("--search-comments", type=str, help="Search comments in a subreddit")
@click.option("--search-post", type=str, help="Search posts in a subreddit")
@click.option(
    "--wiki-page", type=str, help="Get a subreddit's specified wiki page data"
)
@click.option("--wiki-pages", is_flag=True, help="Get a subreddit's wiki pages")
@click.pass_context
def subreddit(
    ctx: click.Context,
    subreddit_name: str,
    comments: bool,
    comments_per_post: int,
    posts: bool,
    profile: bool,
    search_comments: str,
    search_post: str,
    wiki_page: str,
    wiki_pages: bool,
):
    """
    Retrieve data about a specific subreddit including profile, comments, posts, and wiki pages.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param subreddit_name: The name of the subreddit.
    :type subreddit_name: str
    :param comments: Flag to get the subreddit's comments.
    :type comments: bool
    :param comments_per_post: Number of comments per post to retrieve.
    :type comments_per_post: int
    :param posts: Flag to get the subreddit's posts.
    :type posts: bool
    :param profile: Flag to get the subreddit's profile.
    :type profile: bool
    :param search_comments: Query to search comments in the subreddit.
    :type search_comments: str
    :param search_post: Query to search posts in the subreddit.
    :type search_post: str
    :param wiki_page: The name of the wiki page to retrieve.
    :type wiki_page: str
    :param wiki_pages: Flag to get the subreddit's wiki pages.
    :type wiki_pages: bool
    """

    timeframe: reddit.TIMEFRAME = ctx.obj["timeframe"]
    sort: reddit.SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    subreddit_instance = Subreddit(
        name=subreddit_name,
    )
    method_map: Dict = {
        "comments": lambda session, status=None, message=None: subreddit_instance.comments(
            session=session,
            posts_limit=limit,
            comments_per_post=comments_per_post,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
        ),
        "posts": lambda session, status=None, message=None: subreddit_instance.posts(
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
            session=session,
        ),
        "profile": lambda session, status=None, message=None: subreddit_instance.profile(
            status=status, message=message, session=session
        ),
        "search_comments": lambda session, status=None, message=None: subreddit_instance.search_comments(
            query=search_comments,
            posts_limit=limit,
            comments_per_post=comments_per_post,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
            session=session,
        ),
        "search_post": lambda session, status=None, message=None: subreddit_instance.search_posts(
            query=search_post,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
            session=session,
        ),
        "wiki_pages": lambda session, status=None, message=None: subreddit_instance.wiki_pages(
            status=status, message=message, session=session
        ),
        "wiki_page": lambda session, status=None, message=None: subreddit_instance.wiki_page(
            page_name=wiki_page, status=status, message=message, session=session
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            profile=profile,
            comments=comments,
            comments_per_post=comments_per_post,
            posts=posts,
            search_comments=search_comments,
            search_post=search_post,
            wiki_pages=wiki_pages,
            wiki_page=wiki_page,
        )
    )


@cli.command(
    help="Use this command to get all, default, new, and/or popular subreddits.",
)
@click.option("--all", is_flag=True, help="Get all subreddits")
@click.option(
    "--default",
    is_flag=True,
    help="Get default subreddits",
)
@click.option(
    "--new",
    is_flag=True,
    help="Get new subreddits",
)
@click.option(
    "--popular",
    is_flag=True,
    help="Get popular subreddits",
)
@click.pass_context
def subreddits(ctx: click.Context, all: bool, default: bool, new: bool, popular: bool):
    """
    Retrieve various subreddits such as new, popular, default, and all subreddits.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param all: Flag to get all subreddits.
    :type all: bool
    :param default: Flag to get default subreddits.
    :type default: bool
    :param new: Flag to get new subreddits.
    :type new: bool
    :param popular: Flag to get popular subreddits.
    :type popular: bool
    """

    export: str = ctx.obj["export"]
    timeframe: reddit.TIMEFRAME = ctx.obj["timeframe"]
    limit: int = ctx.obj["limit"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    subreddits_instance = Subreddits()
    method_map: Dict = {
        "all": lambda session, status=None, message=None: subreddits_instance.all(
            limit=limit,
            session=session,
            status=status,
            message=message,
        ),
        "default": lambda session, status=None, message=None: subreddits_instance.default(
            limit=limit, session=session, status=status
        ),
        "new": lambda session, status=None, message=None: subreddits_instance.new(
            limit=limit, session=session, status=status
        ),
        "popular": lambda session, status=None, message=None: subreddits_instance.popular(
            limit=limit, session=session, status=status
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            timeframe=timeframe,
            all=all,
            default=default,
            new=new,
            popular=popular,
        )
    )


@cli.command(
    help="Use this command to get user data, such as profile, posts, "
    "comments, top subreddits, moderated subreddits, and more...",
)
@click.argument("username")
@click.option("--comments", is_flag=True, help="Get user's comments")
@click.option(
    "--moderated-subreddits",
    is_flag=True,
    help="Get user's moderated subreddits",
)
@click.option("--overview", is_flag=True, help="Get user's most recent comments")
@click.option("--posts", is_flag=True, help="Get user's posts")
@click.option("--profile", is_flag=True, help="Get user's profile")
@click.option(
    "--search-comments",
    type=str,
    help="Search user's comments that contains a specified query string",
)
@click.option(
    "--search-posts",
    type=str,
    help="Search user's posts that contains a specified query string",
)
@click.option(
    "--top-subreddits",
    type=int,
    help="Get user's top n subreddits",
)
@click.option(
    "--username-available",
    is_flag=True,
    help="Check if the given username is available or taken.",
)
@click.pass_context
def user(
    ctx: click.Context,
    username: str,
    comments: bool,
    moderated_subreddits: bool,
    overview: bool,
    posts: bool,
    profile: bool,
    search_comments: str,
    search_posts: str,
    top_subreddits: int,
    username_available: bool,
):
    """
    Retrieve data about a specific user including profile, posts, comments, and top subreddits.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param username: The username to retrieve data for.
    :type username: str
    :param comments: Flag to get user's comments.
    :type comments: bool
    :param moderated_subreddits: Flag to get user's moderated subreddits.
    :type moderated_subreddits: bool
    :param overview: Flag to get user's most recent comments.
    :type overview: bool
    :param posts: Flag to get user's posts.
    :type posts: bool
    :param profile: Flag to get user's profile.
    :type profile: bool
    :param search_comments: Query to search user's comments.
    :type search_comments: str
    :param search_posts: Query to search user's posts.
    :type search_posts: str
    :param top_subreddits: Number of top subreddits to retrieve.
    :type top_subreddits: int
    :param username_available: Flag to check if the given username is available of taken.
    :type username_available: bool
    """
    timeframe: reddit.TIMEFRAME = ctx.obj["timeframe"]
    sort: reddit.SORT = ctx.obj["sort"]
    limit: int = ctx.obj["limit"]
    export: str = ctx.obj["export"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    user_instance: User = User(
        name=username,
    )
    method_map: Dict = {
        "comments": lambda session, status=None, message=None: user_instance.comments(
            session=session,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
        ),
        "moderated_subreddits": lambda session, status=None, message=None: user_instance.moderated_subreddits(
            session=session, status=status
        ),
        "overview": lambda session, status=None, message=None: user_instance.overview(
            limit=limit, session=session, status=status
        ),
        "posts": lambda session, status=None, message=None: user_instance.posts(
            session=session,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
        ),
        "profile": lambda session, status=None, message=None: user_instance.profile(
            session=session, status=status
        ),
        "search_comments": lambda session, status=None, message=None: user_instance.search_comments(
            query=search_comments,
            limit=limit,
            session=session,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
        ),
        "search_posts": lambda session, status=None, message=None: user_instance.search_posts(
            query=search_posts,
            limit=limit,
            session=session,
            sort=sort,
            timeframe=timeframe,
            status=status,
            message=message,
        ),
        "top_subreddits": lambda session, status=None, message=None: user_instance.top_subreddits(
            session=session,
            top_n=top_subreddits,
            limit=limit,
            timeframe=timeframe,
            status=status,
            message=message,
        ),
        "username_available": lambda session, status=None, message=None: user_instance.username_available(
            session=session, status=status
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            comments=comments,
            moderated_subreddits=moderated_subreddits,
            overview=overview,
            posts=posts,
            profile=profile,
            search_comments=search_comments,
            search_posts=search_posts,
            top_subreddits=top_subreddits,
            username_available=username_available,
        )
    )


@cli.command(
    help="Use this command to get all, new, and/or popular users.",
)
@click.option("--all", is_flag=True, help="Get all users")
@click.option(
    "--new",
    is_flag=True,
    help="Get new users",
)
@click.option(
    "--popular",
    is_flag=True,
    help="Get popular users",
)
@click.pass_context
def users(ctx: click.Context, all: bool, new: bool, popular: bool):
    """
    Retrieve various users such as new, popular, and all users.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param all: Flag to get all users.
    :type all: bool
    :param new: Flag to get new users.
    :type new: bool
    :param popular: Flag to get popular users.
    :type popular: bool
    """

    export: str = ctx.obj["export"]
    timeframe: reddit.TIMEFRAME = ctx.obj["timeframe"]
    limit: int = ctx.obj["limit"]
    time_format: reddit.TIME_FORMAT = ctx.obj["time_format"]

    users_instance = Users()
    method_map: Dict = {
        "all": lambda session, status=None, message=None: users_instance.all(
            session=session, limit=limit, timeframe=timeframe, status=status
        ),
        "new": lambda session, status=None, message=None: users_instance.new(
            session=session, limit=limit, timeframe=timeframe, status=status
        ),
        "popular": lambda session, status=None, message=None: users_instance.popular(
            session=session, limit=limit, timeframe=timeframe, status=status
        ),
    }

    asyncio.run(
        method_call_handler(
            ctx=ctx,
            method_map=method_map,
            export=export,
            all=all,
            new=new,
            popular=popular,
        )
    )


async def call_method(
    method: Callable,
    **kwargs: Union[str, click.Context, aiohttp.ClientSession, Status],
):
    """
    Calls a method with the provided arguments.

    :param method: A method to call.
    :type method: Callable
    :param session: A aiohttp.ClientSession to use for the method's requests.
    :type session: aiohttp.ClientSession
    :param status: An instance of `console.status` used to display animated status messages inside the method.
    :type status: Console.console.status
    :param kwargs: Additional keyword arguments for `export: str`, `argument: str` and `ctx: click.Context` .
    """

    session = kwargs.get("session")
    status = kwargs.get("status")
    message = kwargs.get("message")

    command: str = kwargs.get("ctx").command.name
    argument: str = kwargs.get("argument")

    response_data: Union[List, Dict, str, bool] = await method(
        session=session,
        status=status,
        message=Message,
    )
    if argument == "username_available" and (response_data, bool):
        if response_data:
            Message.ok("Username is available.")
        else:
            Message.warning("Username is already taken.")
    else:
        if response_data:
            dataframe = General.create_dataframe(data=response_data)
            console.print(dataframe)

            if kwargs.get("export"):

                exports_child_dir: str = os.path.join(
                    General.EXPORTS_PARENT_DIR,
                    "exports",
                    command,
                    argument,
                )

                General.pathfinder(
                    directories=[
                        os.path.join(exports_child_dir, extension)
                        for extension in ["csv", "html", "json", "xml"]
                    ],
                )

                export_to: List = kwargs.get("export").split(",")
                General.export_dataframe(
                    dataframe=dataframe,
                    filename=General.filename_timestamp(),
                    directory=exports_child_dir,
                    formats=export_to,
                )


async def method_call_handler(
    ctx: click.Context,
    method_map: Dict,
    export: str,
    **kwargs: Union[str, int, bool],
):
    """
    Handle the method calls based on the provided arguments.

    :param ctx: The Click context object.
    :type ctx: click.Context
    :param method_map: Dictionary mapping method names to their corresponding functions.
    :type method_map: Dict
    :param export: The export format.
    :type export: str
    :param kwargs: Additional keyword arguments.
    :type kwargs: Union[str, int, bool]
    """

    is_valid_arg: bool = False

    for argument, method in method_map.items():
        if kwargs.get(argument):
            is_valid_arg = True
            start_time: datetime = datetime.now()
            try:
                console.print(License.notice, justify="center")
                with Status(
                    status=f"Opening new client session",
                    spinner="dots",
                    spinner_style=Style.yellow.strip("[,]"),
                    console=console,
                ) as status:
                    async with aiohttp.ClientSession() as session:
                        Message.ok("New client session opened")
                        await reddit.infra_status(
                            session=session,
                            status=status,
                            message=Message,
                        )

                        await call_method(
                            method=method,
                            session=session,
                            status=status,
                            ctx=ctx,
                            export=export,
                            argument=argument,
                        )
            except aiohttp.ClientConnectionError as connection_error:
                Message.exception(
                    title="An HTTP error occurred", error=connection_error
                )
            except aiohttp.ClientResponseError as response_error:
                Message.exception(title="An API error occurred", error=response_error)
            finally:
                elapsed_time = datetime.now() - start_time
                Message.ok(
                    f"Session closed. {elapsed_time.total_seconds():.2f} seconds elapsed."
                )

    if not is_valid_arg:
        ctx.command.get_usage(ctx=ctx)


def start():
    """
    Main entrypoint for the Knew Karma command-line interface.
    """

    console.set_window_title(f"{Project.name} {Version.release}")
    cli(obj={})


# -------------------------------- END ----------------------------------------- #



import os
import subprocess
from typing import Optional, List

import aiohttp
from rich.markdown import Markdown
from rich.status import Status

from .general import General
from .terminal import Message, Style

__all__ = ["Package"]


class Package:
    def __init__(self, name: str, **kwargs):

        self._package = name
        self._version = kwargs.get("version")
        self._send_request = kwargs.get("requester")
        self._invalid_package_error = (
            f"The provided package name is not a valid string: {name}"
        )

    async def check_updates(
        self,
        session: aiohttp.ClientSession,
        status: Optional[Status] = None,
        install_if_available: bool = False,
    ):
        """
        Asynchronously checks for updates by comparing the current local version with the remote version.

        Assumes version format: major.minor.patch.prefix

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `Status` object for displaying status messages.
        :type status: Optional[rich.status.Status]
        """

        if status:
            status.update("Checking for updates")

        # Make a GET request to GitHub to get the project's latest release.
        response = await self._send_request(
            endpoint=f"https://api.github.com/repos/knewkarma-io/{self._package}/releases/latest",
            session=session,
        )

        if response.get("tag_name"):
            remote_version_str = response.get("tag_name")
            markup_release_notes = response.get("body")

            # Splitting the version strings into components
            local_version_parts = [
                self._version.major[0],
                self._version.minor[0],
                self._version.patch[0],
            ]
            remote_version_parts: List = remote_version_str.split(".")

            local_major: int = local_version_parts[0]
            local_minor: int = local_version_parts[1]
            local_patch: int = local_version_parts[2]

            remote_major = int(remote_version_parts[0])
            remote_minor = int(remote_version_parts[1])
            remote_patch = int(remote_version_parts[2])

            update_level = ""

            # Check for differences in version parts
            if remote_major != local_major:
                update_level = f"{Style.red}{self._version.major[1]}{Style.reset}"

            elif remote_minor != local_minor:
                update_level = f"{Style.yellow}{self._version.minor[1]}{Style.reset}"

            elif remote_patch != local_patch:
                update_level = f"{Style.green}{self._version.patch[1]}{Style.reset}"

            if update_level:
                markdown_release_notes = Markdown(markup=markup_release_notes)
                General.make_panel(
                    title=f"{Style.bold}{update_level} Update Available ({Style.cyan}{remote_version_str}{Style.reset}){Style.reset}",
                    content=markdown_release_notes,
                    subtitle=f"{Style.bold}{Style.italic}Thank you, for using Knew Karma!{Style.reset}{Style.reset} ❤️ ",
                )

                if install_if_available:
                    self.update_package(status=status)
            else:
                Message.ok(message=f"No new releases")

    @staticmethod
    def is_docker_container() -> bool:
        """
        Determines if the program is running inside a Docker container.

        :return: True if running inside a Docker container, otherwise False.
        :rtype: bool
        """
        return os.environ.get("IS_DOCKER_CONTAINER") == "1"

    def is_snap_package(self) -> bool:
        """
        Checks if a specified package name is installed as a snap package
        by checking if it's running inside a SNAP environment.

        :return: True if the specified package is installed as a snap package, otherwise False.
        :rtype: bool
        """

        if self._package:
            return True if os.getenv("SNAP") else False
        else:
            Message.error(self._invalid_package_error)

    def is_pypi_package(self) -> bool:
        """
        Checks if a specified package name is installed as a pypi package.

        :return: True if the specified package is installed as a pypi package, otherwise False.
        :rtype: bool
        """

        if self._package:
            try:
                __import__(name=self._package)
                return True
            except ImportError:
                return False
        else:
            Message.error(self._invalid_package_error)

    def update_package(self, status: Optional[Status] = None):
        """
        Updates a specified pypi package.

        :param status: An optional `Status` object for displaying status messages.
        :type status: rich.status.Status
        """

        if self.is_snap_package():
            Message.warning("Run 'snap refresh knewkarma' to update the snap package.")
        else:
            if self._package:
                try:
                    if status:
                        status.stop()
                        # status.update("Downloading updates. Please wait")
                    subprocess.run(
                        [
                            "pip",
                            "install",
                            "--upgrade",
                            f"{self._package}[core]",
                        ],
                        check=True,
                    )
                    Message.ok(f"DONE. The updates will be installed on next run.")
                except subprocess.CalledProcessError as called_process_error:
                    Message.exception(
                        title=f"An error occurred ({Style.italic}while updating package{Style.reset})",
                        error=called_process_error,
                    )
                except Exception as unexpected_error:
                    Message.exception(unexpected_error)
            else:
                Message.error(self._invalid_package_error)


# -------------------------------- END ----------------------------------------- #



__all__ = ["Author", "Project"]


class Author:
    name: str = "Richard Mwewa"
    username: str = "rly0nheart"
    gravatar: str = f"https://gravatar.com/{username}"


class Project:
    name: str = "Knew Karma"
    package: str = "knewkarma"
    opencollective: str = f"https://opencollective.com/{package}"
    documentation: str = f"https://{package}.readthedocs.io"
    summary: str = f"A Reddit data analysis toolkit — by {Author.name}"
    description: str = f"""
{name} (/nuː ‘kɑːrmə/) is a Reddit data analysis toolkit designed to provide an extensive range of
functionalities for exploring and analysing Reddit data. It includes a Command-Line Interface (CLI), and an
Application Programming Interface (API) to enable an easy integration in other Python Projects.
"""


# -------------------------------- END ----------------------------------------- #



import os
from datetime import datetime
from types import SimpleNamespace
from typing import Union, Literal, List, Tuple, Dict, Optional

import pandas as pd
from rich.box import DOUBLE
from rich.console import ConsoleRenderable, RichCast
from rich.panel import Panel

from .terminal import console, Message

__all__ = ["General"]


class General:

    EXPORT_FORMATS = Literal["csv", "html", "json", "xml"]
    EXPORTS_PARENT_DIR: str = os.path.expanduser(os.path.join("~", "knewkarma"))

    @staticmethod
    def is_matplotlib_installed() -> bool:
        try:
            __import__("matplotlib.pyplot")

            is_installed: bool = True
        except ImportError:
            is_installed = False

        return is_installed

    def plot_bar_chart(
        self,
        data: Dict[str, int],
        title: str,
        xlabel: str,
        ylabel: str,
        colours: List[str],
        filename: str,
        figure_size: Tuple[int, int] = (10, 5),
    ):
        """
        Plots a bar chart for the given data.

        :param data: A dictionary where keys are the categories and values are the counts or frequencies.
        :type data: list[str, int]
        :param title: The title of the plot.
        :type title: str
        :param xlabel: The label for the x-axis.
        :type xlabel: str
        :param ylabel: The label for the y-axis.
        :type ylabel: str
        :param colours: A list of colours to use for the bars in the chart.
        :type colours: list[str]
        :param filename: The name of the file where the plot will be saved.
        :type filename: str
        :param figure_size: The size of the figure (width, height). Defaults to (10, 5).
        :type figure_size: tuple[int, int]
        """

        if self.is_matplotlib_installed():
            import matplotlib.pyplot as plt

            plt.figure(figsize=figure_size)
            plt.bar(list(data.keys()), list(data.values()), color=colours)
            plt.title(title)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.savefig(f"{filename}.png")

            Message.ok(f"{title} saved to [link file://{filename}.png]{filename}.png")

    @staticmethod
    def create_dataframe(
        data: Union[SimpleNamespace, List[SimpleNamespace], List[Tuple[str, int]]],
    ) -> pd.DataFrame:
        """
        Makes a Pandas dataframe from the provided data.

        :param data: Data to be converted.
        :type data: Union[SimpleNamespace, List[SimpleNamespace], List[Tuple[str, int]]]
        :return: A pandas DataFrame constructed from the provided data. Excludes any 'raw_data'
                 column from the dataframe.
        :rtype: pd.DataFrame
        """

        if isinstance(data, SimpleNamespace):
            # Transform each attribute of the object into a dictionary entry
            transformed_data = [
                {"key": key, "value": value}
                for key, value in data.data.__dict__.items()
            ]

        # Convert a list of SimpleNamespace objects to a list of dictionaries
        elif isinstance(data, List) and all(
            isinstance(item, SimpleNamespace) for item in data
        ):
            # Each object in the list is transformed to its dictionary representation
            transformed_data = [item.data.__dict__ for item in data]
        else:
            transformed_data = data

        # Set pandas display option to show all rows
        pd.set_option("display.max_rows", None)

        # Create a DataFrame from the transformed data
        dataframe = pd.DataFrame(transformed_data)

        return dataframe.dropna(axis=1, how="all")

    def export_dataframe(
        self,
        dataframe: pd.DataFrame,
        filename: str,
        directory: str,
        formats: List[EXPORT_FORMATS],
    ):
        """
        Exports a Pandas dataframe to specified file formats.

        :param dataframe: Pandas dataframe to export.
        :type dataframe: pandas.DataFrame
        :param filename: Name of the file to which the dataframe will be exported.
        :type filename: str
        :param directory: Directory to which the dataframe files will be saved.
        :type directory: str
        :param formats: A list of file formats to which the data will be exported.
        :type formats: List[Literal]
        """

        file_mapping: Dict = {
            "csv": lambda: dataframe.to_csv(
                os.path.join(directory, "csv", f"{filename}.csv"), encoding="utf-8"
            ),
            "html": lambda: dataframe.to_html(
                os.path.join(directory, "html", f"{filename}.html"),
                escape=False,
                encoding="utf-8",
            ),
            "json": lambda: dataframe.to_json(
                os.path.join(directory, "json", f"{filename}.json"),
                force_ascii=False,
                indent=4,
            ),
            "xml": lambda: dataframe.to_xml(
                os.path.join(directory, "xml", f"{filename}.xml"),
                parser="etree",
                encoding="utf-8",
            ),
        }

        for file_format in formats:
            if file_format in file_mapping:
                filepath: str = os.path.join(
                    directory, file_format, f"{filename}.{file_format}"
                )
                file_mapping.get(file_format)()
                Message.ok(
                    f"{self.get_file_size(file_path=filepath)} written to [link file://{filepath}]{filepath}"
                )
            else:
                continue

    @staticmethod
    def get_file_size(file_path: str) -> str:
        """
        Gets a file size and puts it in human-readable form.

        :param file_path: Path to target file.
        :type file_path: str
        :return: A human-readable form of the file size.
        :rtype: str
        """

        file_size_bytes: int = os.path.getsize(file_path)
        units: list = ["B", "KB", "MB", "GB", "TB", "PB"]

        unit_index: int = 0

        while file_size_bytes >= 1024 and unit_index < len(units) - 1:
            file_size_bytes /= 1024
            unit_index += 1

        return f"{file_size_bytes:.2f} {units[unit_index]}"

    @staticmethod
    def make_panel(
        title: str,
        content: Union[ConsoleRenderable, RichCast, str],
        subtitle: Optional[str] = None,
    ):
        """
        Makes a rich Panel for whatever data is needed to be placed in it.

        :param title: Panel title.
        :type title: str
        :param content: Content of the panel.
        :type content: Union[rich.ConsoleRenderable, rich.RichCast, str]
        :param subtitle: Panel subtitle.
        :type subtitle: str
        """
        from .terminal import Style

        console.print(
            Panel(
                renderable=content,
                title=f"{Style.bold}{title}{Style.reset}",
                subtitle=(subtitle if subtitle else None),
                box=DOUBLE,
                style=f"{Style.white.strip('[,]')} on black",
            )
        )

    @staticmethod
    def filename_timestamp() -> str:
        """
        Generates a timestamp string suitable for file naming, based on the current date and time.
        The format of the timestamp is adapted based on the operating system.

        :return: The formatted timestamp as a string. The format is "%d-%B-%Y-%I-%M-%S%p" for Windows
                 and "%d-%B-%Y-%I:%M:%S%p" for non-Windows systems.
        :rtype: str

        Example
        -------
        - Windows: "20-July-1969-08-17-45PM"
        - Non-Windows: "20-July-1969-08:17:45PM"
        """
        now = datetime.now()
        return (
            now.strftime("%d-%B-%Y-%I-%M-%S%p")
            if os.name == "nt"
            else now.strftime("%d-%B-%Y-%I:%M:%S%p")
        )

    @staticmethod
    def pathfinder(directories: Union[List[str], str]):
        """
        Creates directories for exported data (`exported`).

        :param directories: A list of directories or a directory name to create
        :type directories: Union[List[str], str]
        :raise Exception: If any was encountered.
        """

        try:
            if isinstance(directories, List) and all(
                isinstance(directory, str) for directory in directories
            ):
                for directory in directories:
                    os.makedirs(name=directory, exist_ok=True)
            elif isinstance(directories, str):
                os.makedirs(name=directories, exist_ok=True)
        except Exception as unexpected_error:
            Message.exception(unexpected_error)


# -------------------------------- END ----------------------------------------- #



from rich.console import Console

__all__ = ["console", "Message", "Style"]

console = Console(log_time=False)


class Style:
    """
    Provides rich formatting codes for styling console text.

    Attributes:
        blue (str): Rich code for blue text.
        bold (str): Rich code for bold text.
        cyan (str): Rich code for cyan text.
        green (str): Rich code for green text.
        italic (str): Rich code for italic text.
        yellow (str): Rich code for yellow text.
        purple (str): Rich code for purple text.
        red (str): Rich code for red text.
        white (str): Rich code for white text.
        reset (str): Rich code to reset text formatting.
    """

    blue: str = "[blue]"
    bold: str = "[bold]"
    cyan: str = "[cyan]"
    green: str = "[green]"
    italic: str = "[italic]"
    yellow: str = "[yellow]"
    purple: str = "[purple]"
    red: str = "[red]"
    white: str = "[white]"
    reset: str = "[/]"


class Message:
    """
    Provides static methods for printing and/or logging formatted notifications to the console.
    """

    @staticmethod
    def ok(message: str):
        """
        Prints a `success` message (with a checkmark).

        :param message: Text to be logged.
        :type message: str
        """

        console.print(f"{Style.green}✔{Style.reset} {message}")

    @staticmethod
    def info(message: str):
        """
        Prints an `informational` message to the console.

        :param message: Text to be logged.
        :type message: str
        """

        console.print(f"{Style.green}✱{Style.reset} {message}")

    @staticmethod
    def warning(message: str):
        """
        Prints a `warning` message to the console.

        :param message: Text to be logged.
        :type text: str
        """

        console.print(f"{Style.yellow}✘{Style.reset} {message}")

    @staticmethod
    def error(text: str):
        """
        Logs an error with the specified message to the console.

        :param text: Error message to be logged.
        :type text: str
        """

        console.log(f"{Style.yellow}✘{Style.reset} {message}")

    @staticmethod
    def exception(error: Exception, title: str = "An unexpected error occurred"):
        """
        Logs an exception message to the console.

        :param title: Title of the error (e.g., "An exception occurred")
        :type title: str
        :param error: Error that triggered the exception.
        :type error: Exception
        """

        console.log(f"{Style.red}✘{Style.reset} {title}: {error}")


# -------------------------------- END ----------------------------------------- #



class Version:
    major: tuple[int, str] = 7, "MAJOR"
    minor: tuple[int, str] = 1, "MINOR"
    patch: tuple[int, str] = 2, "PATCH"
    full_version: str = f"{major[0]}.{minor[0]}.{patch[0]}"
    release: str = f"{major[0]}.{minor[0]}"



from collections import Counter
from platform import python_version, platform
from types import SimpleNamespace
from typing import Literal, Union, Optional, List

import aiohttp
import kraw

from .meta.about import Project
from .meta.version import Version
from .utils.general import General

__all__ = [
    "reddit",
    "Post",
    "Posts",
    "Search",
    "Subreddit",
    "Subreddits",
    "User",
    "Users",
]


reddit = kraw.Reddit(
    headers={
        "User-Agent": f"{Project.name.replace(' ', '-')}/{Version.release} "
        f"(Python {python_version} on {platform}; +{Project.documentation})"
    },
)


class Comment:
    """Represents a Reddit comment and provides methods for interacting with it."""

    pass


class Post:
    """Represents a Reddit post and provides methods for fetching post data and comments."""

    def __init__(self, id: str, subreddit: str):
        """
        Initialises a `Post` instance to fetch the post's data and comments.

        :param id: The ID of the Reddit post to retrieve.
        :type id: str
        :param subreddit: The subreddit where the post was created.
        :type subreddit: str
        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """

        self._id = id
        self._subreddit = subreddit

    async def data(
        self,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> SimpleNamespace:
        """
        Asynchronously retrieves data for a Reddit post, excluding comments.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A `SimpleNamespace` object containing parsed post data.
        :rtype: SimpleNamespace
        """

        post_data = await reddit.post(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            status=status,
            id=self._id,
            subreddit=self._subreddit,
        )

        return post_data

    async def comments(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves comments for a Reddit post.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of comments to retrieve.
        :type limit: int
        :param sort: The sorting criterion for the comments. Defaults to "all".
        :type sort: SORT, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed comment data.
        :rtype: List[SimpleNamespace]
        """

        comments_data = await reddit.comments(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="post",
            id=self._id,
            subreddit=self._subreddit,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        return comments_data


class Posts:
    """Represents Reddit posts and provides methods for retrieving posts from various sources."""

    @staticmethod
    async def best(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the best posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        best_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="best",
            limit=limit,
            timeframe=timeframe,
        )

        return best_posts

    @staticmethod
    async def controversial(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the controversial posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        controversial_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="controversial",
            limit=limit,
            timeframe=timeframe,
        )

        return controversial_posts

    @staticmethod
    async def front_page(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        sort: reddit.SORT = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the front-page posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param sort: Sorting criterion for posts. Defaults to "all".
        :type sort: SORT, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        front_page_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="front_page",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        return front_page_posts

    @staticmethod
    async def new(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        sort: reddit.SORT = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the new posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param sort: Sorting criterion for posts. Defaults to "all".
        :type sort: SORT, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        new_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="new",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        return new_posts

    @staticmethod
    async def popular(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the popular posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """
        popular_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="popular",
            limit=limit,
            timeframe=timeframe,
        )

        return popular_posts

    @staticmethod
    async def rising(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves the rising posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        rising_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="rising",
            limit=limit,
            timeframe=timeframe,
        )

        return rising_posts


class Search:
    """
    Represents the Reddit search functionality and provides methods for retrieving search results
    from different entities.
    """

    def __init__(self, query: str):
        """
        Initialises the `Search` instance for searching posts, subreddits, and users.

        :param query: The search query string.
        :type query: str
        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """

        self._query = query

    async def posts(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves posts that match with the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param sort: Sorting criterion for posts. Defaults to "all".
        :type sort: SORT, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await reddit.search(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="posts",
            query=self._query,
            sort=sort,
            limit=limit,
        )

        return search_results

    async def subreddits(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves subreddits that match with the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to retrieve.
        :type limit: int
        :param sort: Sorting criterion for subreddits. Defaults to "all".
        :type sort: SORT, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed subreddit data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await reddit.search(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="subreddits",
            query=self._query,
            sort=sort,
            limit=limit,
        )

        return search_results

    async def users(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves users that match with the specified query.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of users to retrieve.
        :type limit: int
        :param sort: Sorting criterion for users. Defaults to "all".
        :type sort: SORT, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed user data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await reddit.search(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="users",
            query=self._query,
            sort=sort,
            limit=limit,
        )

        return search_results


class Subreddit:
    """Represents a Reddit community (subreddit) and provides methods for retrieving its data."""

    def __init__(self, name: str):
        """
        Initialises a `Subreddit` instance to get a subreddit's profile, wiki pages, and posts data,
        and to search for posts containing a specified query.

        :param name: The name of the subreddit to retrieve data from.
        :type name: str
        :param time_format: Format for displaying time, either 'concise' or 'locale'. Defaults to 'locale'.
        :type time_format: Literal["concise", "locale"]
        """

        self._name = name

    async def comments(
        self,
        session: aiohttp.ClientSession,
        posts_limit: int,
        comments_per_post: int,
        sort: reddit.SORT = "all",
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves comments from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param posts_limit: Maximum number of posts to retrieve comments from.
        :type posts_limit: int
        :param comments_per_post: Maximum number of comments to retrieve per post.
        :type comments_per_post: int
        :param sort: Sorting criterion for the posts and comments. Defaults to "all".
        :type sort: SORT, optional
        :param timeframe: The timeframe from which to retrieve posts and comments. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed comment data.
        :rtype: List[SimpleNamespace]
        """

        posts = await self.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            limit=posts_limit,
            sort=sort,
            timeframe=timeframe,
        )
        all_comments: List = []
        for post in posts:
            post = Post(
                id=post.get("id"),
                subreddit=post.get("subreddit"),
            )
            post_comments = await post.comments(
                session=session,
                proxy=proxy,
                proxy_auth=proxy_auth,
                limit=comments_per_post,
                sort=sort,
                status=status,
            )

            all_comments.extend(post_comments)

        return all_comments

    async def posts(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously retrieves posts from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to retrieve.
        :type limit: int
        :param sort: Sorting criterion for the posts. Defaults to "all".
        :type sort: SORT, optional
        :param timeframe: The timeframe from which to retrieve posts. Defaults to "all".
        :type timeframe: reddit.TIMEFRAME, optional
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing parsed post data.
        :rtype: List[SimpleNamespace]
        """

        subreddit_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="subreddit",
            subreddit=self._name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        return subreddit_posts

    async def profile(
        self,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> SimpleNamespace:
        """
        Asynchronously retrieves a subreddit's profile data.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A `SimpleNamespace` object containing the parsed subreddit profile data.
        :rtype: SimpleNamespace
        """

        subreddit_profile = await reddit.subreddit(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            status=status,
            name=self._name,
        )

        return subreddit_profile

    async def search(
        self,
        session: aiohttp.ClientSession,
        query: str,
        limit: int,
        sort: reddit.SORT = "all",
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get posts that contain the specified query string from a subreddit.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param query: Search query.
        :type query: str
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing post data.
        :rtype: List[SimpleNamespace]
        """

        search_results = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="search_subreddit",
            subreddit=self._name,
            query=query,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        return search_results

    async def wikipages(
        self,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[str]:
        """
        Asynchronously get a subreddit's wiki pages.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of strings, each representing a wiki page.
        :rtype: List[str]
        """

        if status:
            status.update(
                f"Retrieving wiki pages from subreddit ({self._name})",
            )

        pages = await reddit.connection.send_request(
            endpoint=f"{reddit.connection.endpoints.subreddit}/{self._name}/wiki/pages.json",
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
        )

        return pages.get("data")

    async def wikipage(
        self,
        page_name: str,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> SimpleNamespace:
        """
        Asynchronously get a subreddit's specified wiki page data.

        :param page_name: Wiki page to get data from.
        :type page_name: str
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A `SimpleNamespace` object containing the parsed wiki page data.
        :rtype: SimpleNamespace
        """

        wiki_page = await reddit.wiki_page(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            status=status,
            name=page_name,
            subreddit=self._name,
        )

        return wiki_page


class Subreddits:
    """Represents Reddit subreddits and provides methods for getting related data."""

    def __init__(self):
        """
        Initialises the `Subreddits()` instance for getting `all`, `default`, `new` and `popular` subreddits.

        :param time_format: Time format of the output data. Use `concise` for a human-readable
                        time difference, or `locale` for a localized datetime string. Defaults to `locale`.
        :type time_format: Literal["concise", "locale"]
        """

    async def all(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get all subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get all subreddits.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]

        Note:
            Items will most likely be limited to 1000, per Reddit's API policy.
        """

        all_subreddits = await reddit.subreddits(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="all",
            limit=limit,
            timeframe=timeframe,
        )

        return all_subreddits

    async def default(
        self,
        limit: int,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get default subreddits.

        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """

        default_subreddits = await reddit.subreddits(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="default",
            timeframe="all",
            limit=limit,
        )

        return default_subreddits

    async def new(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get new subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get new subreddits.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """
        new_subreddits = await reddit.subreddits(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="new",
            limit=limit,
            timeframe=timeframe,
        )

        return new_subreddits

    async def popular(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get popular subreddits.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param timeframe: Timeframe from which to get popular subreddits.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """

        popular_subreddits = await reddit.subreddits(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="popular",
            limit=limit,
            timeframe=timeframe,
        )

        return popular_subreddits


class User:
    """Represents a Reddit user and provides methods for getting data from the specified user."""

    def __init__(self, name: str):
        """
        Initialises a `User()` instance for getting a user's `profile`, `posts` and `comments` data.

        :param name: Username to get data from.
        :type name: str
        :param time_format: Time format of the output data. Use "concise" for a human-readable
                        time difference, or "locale" for a localized datetime string. Defaults to "locale".
        :type time_format: Literal["concise", "locale"]
        """

        self._name = name

    async def comments(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's comments.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of comments to return.
        :type limit: int
        :param sort: Sort criterion for the comments.
        :type sort: str
        :param timeframe: Timeframe from which tyo get comments.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing comment data.
        :rtype: List[SimpleNamespace]
        """

        user_comments = await reddit.comments(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="user",
            username=self._name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
        )

        return user_comments

    async def moderated_subreddits(
        self,
        session: aiohttp.ClientSession,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get subreddits moderated by user.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing subreddit data.
        :rtype: List[SimpleNamespace]
        """

        subreddits = await reddit.subreddits(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="user_moderated",
            username=self._name,
            limit=0,
        )

        return subreddits

    async def overview(
        self,
        limit: int,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's most recent comments.

        :param limit: Maximum number of comments to return.
        :type limit: int
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing data about a recent comment.
        :rtype: List[SimpleNamespace]
        """

        user_overview = await reddit.comments(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="user_overview",
            limit=limit,
            timeframe="all",
            username=self._name,
        )

        return user_overview

    async def posts(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: reddit.SORT = "all",
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get a user's posts.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing post data.
        :rtype: List[SimpleNamespace]
        """

        user_posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="user",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            username=self._name,
        )

        return user_posts

    async def profile(
        self,
        session: aiohttp.ClientSession,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> SimpleNamespace:
        """
        Asynchronously get a user's profile data.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A SimpleNamespace object containing user profile data.
        :rtype: SimpleNamespace
        """

        user_profile = await reddit.user(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            status=status,
            name=self._name,
        )

        return user_profile

    async def top_subreddits(
        self,
        session: aiohttp.ClientSession,
        top_n: int,
        limit: int,
        filename: str = None,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> Union[List[tuple[str, int]], None]:
        """
        Asynchronously get a user's top n subreddits based on subreddit frequency in n posts and saves the analysis to a file.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param top_n: Communities arranging number.
        :type top_n: int
        :param limit: Maximum number of posts to scrape.
        :type limit: int
        :param filename: Filename to which the analysis will be saved.
        :type filename: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        """

        posts = await reddit.posts(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="user",
            limit=limit,
            timeframe=timeframe,
            username=self._name,
        )

        if posts:
            # Extract subreddit names
            subreddits = [post.get("data", {}).get("subreddit") for post in posts]

            # Count the occurrences of each subreddit
            subreddit_counts: Counter = Counter(subreddits)

            # Get the most common subreddits
            top_subreddits: List[tuple[str, int]] = subreddit_counts.most_common(top_n)

            # Prepare data for plotting
            subreddit_names = [subreddit[0] for subreddit in top_subreddits]
            subreddit_frequencies = [subreddit[1] for subreddit in top_subreddits]

            if General.is_matplotlib_installed():
                General.plot_bar_chart(
                    data=dict(zip(subreddit_names, subreddit_frequencies)),
                    title=f"top {top_n}/{limit} subreddits analysis",
                    xlabel="Subreddits",
                    ylabel="Frequency",
                    figure_size=(top_n + 10, 5),
                    colours=["lightblue"] * top_n,
                    filename=f"{filename + '_' if filename else ''}top_{top_n}_of_{limit}_subreddits",
                )
            else:
                return top_subreddits

    async def username_available(
        self,
        session: aiohttp.ClientSession,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> bool:
        """
        Checks if the given username is available or taken.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: `True` if the given username is available. Otherwise, `False`.
        :rtype: bool
        """

        if status:
            status.update(f"Checking username availability: {self._name}")

        response: bool = await reddit.connection.send_request(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            endpoint=reddit.connection.endpoints.username_available,
            params={"user": self._name},
        )

        return response


class Users:
    """Represents Reddit users and provides methods for getting related data."""

    @staticmethod
    async def new(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get new users.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of new users to return.
        :type limit: int
        :param timeframe: Timeframe from which to get new posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing a user's data.
        :rtype: List[SimpleNamespace]
        """

        new_users = await reddit.users(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="new",
            limit=limit,
            timeframe=timeframe,
        )

        return new_users

    @staticmethod
    async def popular(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get popular users.

        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of popular users to return.
        :type limit: int
        :param timeframe: Timeframe from which to get popular posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing a user's data.
        :rtype: List[SimpleNamespace]
        """

        popular_users = await reddit.users(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="popular",
            limit=limit,
            timeframe=timeframe,
        )

        return popular_users

    @staticmethod
    async def all(
        session: aiohttp.ClientSession,
        limit: int,
        timeframe: reddit.TIMEFRAME = "all",
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        status: Optional[kraw.dummies.Status] = kraw.dummies.Status,
        message: Optional[kraw.dummies.Message] = kraw.dummies.Message,
    ) -> List[SimpleNamespace]:
        """
        Asynchronously get all users.

        :param limit: Maximum number of all users to return.
        :type limit: int
        :param session: An `aiohttp.ClientSession` for making the HTTP request.
        :type session: aiohttp.ClientSession
        :param timeframe: Timeframe from which to get all posts.
        :type timeframe: Literal[str]
        :param status: An optional `rich.status.Status` object for displaying status messages. Defaults to None.
        :type status: Optional[rich.status.Status]
        :return: A list of `SimpleNamespace` objects, each containing a user's data.
        :rtype: List[SimpleNamespace]
        """

        all_users = await reddit.users(
            session=session,
            proxy=proxy,
            proxy_auth=proxy_auth,
            message=message,
            status=status,
            kind="all",
            limit=limit,
            timeframe=timeframe,
        )

        return all_users


# -------------------------------- END ----------------------------------------- #
